// Copyright (C) 1999 Joshua Fox

package ddj;

import com.sun.media.jsdt.Channel;
import com.sun.media.jsdt.Data;
import com.sun.media.jsdt.JSDTException;
import com.sun.media.jsdt.NameInUseException;
import com.sun.media.jsdt.PortInUseException;
import com.sun.media.jsdt.RegistryFactory;
import com.sun.media.jsdt.NoRegistryException;
import com.sun.media.jsdt.NoSuchSessionException;
import com.sun.media.jsdt.Session;
import com.sun.media.jsdt.SessionFactory;
import com.sun.media.jsdt.URLString;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * The main class for the Application. Run it with
 *   <CODE>java ChatApplication registryServerName</CODE>
 */
public class ChatApplication {
    /**
     * The registy is the a central lookup point in the JSDT.
     * You must the registry server address on the command line.
     * If you write your JSDT client application in an applet
     * (but note the difficulties with this mentioned
     * in the article), then you can get the server name
     * without need for outside configuration by
     * using applet.getCodebase.getHost()
     */
    private String registryServerName;

    /**
     * The current chat session.
     */
    private Session currentSession;

    /**
     * The name of the JSDT Client in the JSDT system.
     */
    private String clientName;


    /**
     * The JSDT Client object represents this application in
     * the JSDT system; in particular, it provides a name
     * identifying this application in the system.
     */
    private JSDTClient jsdtClient;

    /**
     * Channel for chatting.
     */
    private Channel chatChannel;

    /**
     * The user interface.
     */
    private ChatApplicationGUI chatGui;

    /**
     * Listen to who enters and leaves the Session.
     */
    private ChatSessionListener chatSessionListener;

    /**
     * The Client who initiates a Session has to set
     * up a ServerSocket and so has a special
     * status in the Session. When it leaves, the
     * Session is no longer useable.
     */
    private boolean isSessionInitiator = false;


    /**
     *  In the constructuor, we set up the components
     * used by this application. The comments below
     * will explain them one by one.
     */
    public ChatApplication(String registryServer) {
        // The Registry Server is the central
        // lookup point in a JSDT system.
        registryServerName = registryServer;

        // The Graphical User Interface 
        chatGui = new ChatApplicationGUI(this);
    }


    /**
     * Connect to the indicated Session. Must set clientName and
     * sessionName in GUI before calling this method.
     */
    public synchronized void connectToSession() throws ChatException {
        String clientName = chatGui.getClientName();
        String sessionName = chatGui.getSessionName();

        if (clientName == null || clientName.trim().equals("")){
            throw new ChatException("Must set Client name to connect to Session");
        }

        if (sessionName ==null || sessionName.trim().equals("")){
           throw new ChatException("Must set Session name to connect to Session");
        }

        //clean up any existing Session
        if (currentSession != null) {
            disconnect();
        }

        // create a Client object if necessary
        if (jsdtClient == null) {
            // The Client represents this application in the JSDT system.
            jsdtClient = new JSDTClient(clientName);
        }

        try {
            // JSDT uses URLString objects rather than java.net.URL
            // to designate its URLS (which begin with "jsdt://")
            URLString sessionUrlString = URLString.createSessionURL(registryServerName,
                                                    Constants.SESSION_PORT,
                                                    Constants.JSDT_IMPLEMENTATION,
                                                    sessionName);

            // The Client application does not create Sessions;
            // it just gets a reference to an existing Session.
            if (!SessionFactory.sessionExists(sessionUrlString)) {
                throw new ChatException("Trying to get a reference to URLString " +
                            sessionUrlString + "; No such Session exists.");
            }

            // The method createSession() creates the Session if
            // it does not exist, or gets a reference to it if
            // it does exist. In the client application, we will
            // not actually create the Session, just get a reference.
            // The "true" parameter in createSesssion() indicates
            // that we should not only get a reference to a Session,
            // but also join it.
            //
            // createSession() is where you are likely to get the most ,
            // exceptionssince this is the method which actually connects
            //  you (using a socket) to the JSDT system. If there is
            // such as a broken or clogged network connection or a firewall,
            // anything wrong, you will first learn about it here.
            currentSession = SessionFactory.createSession(jsdtClient,
                                                          sessionUrlString,
                                                          true);

            // Listen to the Session and update the GUI
            // when someone enters or leaves.
            chatSessionListener = new ChatSessionListener(chatGui, this);
            currentSession.addSessionListener(chatSessionListener);

            // createChannel() creates a Channel within the Session
            // if the Channel does not exist, or just gets a
            // reference to it if it does exist.
            // Explanation of parameters:
            // The first parameter is the Client that is joining the
            //      Channel. This client would be authenticated by the
            //      ChannelManager, if there were one.
            // The second parameter is the name of the Channel.
            // The third is the reliability--as in the difference
            //      between TCP and UDP.
            // The fourth is true if you want to ensure that the Data
            //      arrive in the order in which they were sent.
            // The last parameter says that the Client should join the
            //      Channel as part of this method call, rather than just getting
            //      a reference to the Channel here.
            chatChannel = currentSession.createChannel(jsdtClient,
                                                    Constants.CHAT_CHANNEL_NAME,
                                                    true, true, true);

            // The Channel Consumer is the call-back interface for
            // asynchronously receiving data that arrives in the JSDT Channel.
            chatChannel.addConsumer(jsdtClient, new ChatConsumer(chatGui));

            // Update the GUI about what Clients are now in this apps Session.
            chatGui.setClientsInSession(currentSession.listClientNames());

        } catch (NameInUseException niue){
            jsdtClient = null;
            throw new ChatException("In connectToSession()", niue);
        } catch (NoRegistryException nre) {
            System.err.println("No registry is available on server " + registryServerName);
        }  catch (Exception e) {
             throw new ChatException("In connectToSession()", e);
        }

   }


    /**
     * Send the messsage indicated in the parameter. Not synchronized,
     * since the underlying Channel sendX() methods are synchronized. 
     */
    public void sendMessage(String message) throws ChatException{
        try {
            Data data = new Data(message);
            if (chatChannel == null) {
                throw new ChatException("Channel not yet initialized ");
            }

            // This is the actual sending method. It comes in three variants,
            // sendToAll: Everyone in the Channel, including this Client.
            // sendToOthers: Everyone in the Channel, except for this Client.
            // sendToClient: Send to the Client of the name given
            //          in the parameter.
            chatChannel.sendToAll(jsdtClient, data);

        } catch (JSDTException jsdte) {
            jsdte.printStackTrace();
            disconnect();
            throw new ChatException("In sendMessage(" + message + ")", jsdte);
        }
    }


    /**
     * Disconnect from the current session.
     */
    public synchronized void disconnect() {
        chatGui.setEndSession();

        if (currentSession == null) {
            return;
        }

        try {
           currentSession.removeSessionListener(chatSessionListener);
        } catch (JSDTException jsdte) {
            System.err.println("Exception in removing Session Listener; " +  jsdte);
        }

        try {
            // session.close(false) closes only the connection associated
            // with this client. You will want to use the false value
            // to avoid closing other connections in the virtual machine.
            // You would use true when you really want to clean up-- say,
            // right before you call System.exit().
            currentSession.close(false);
            currentSession = null;
        } catch (Exception e) {
            // In certain cases, when you disconnect after the server
            // is closed , the JSDT will throw NullPointerException
            // rather than JSDTException, so here we catch a general
            //Exception
            System.err.println("Exception in closing Session; " +  e);
        }
    }

    /**
     *  @return the Session we are now in, or "" if we are not in a Session
     */
    public String  getCurrentSession() {
        if (currentSession != null) {
            return currentSession.getName();
        } else {
            return "";
        }
    }

    /**
     *
     */
     public void setClientName(String name ) {
        clientName = name;
     }

     /**
      * @return an array of all current Sessions, or null if an exception occurs.
      */
     public String [] getCurrentSessions() {
        try {

            URLString [] sessionUrls = SessionFactory.listSessions(registryServerName,
                                          Constants.JSDT_IMPLEMENTATION);

            String [] sessions = new String[sessionUrls.length];
            for (int i = 0; i < sessions.length; i++) {
                sessions[i] = sessionUrls[i].getObjectName();
            }

            return sessions;
        } catch (JSDTException jsdte) {
             return null;
        }
     }


     /**
      *  @param args, where args[0] is the name of the registry server.
      *
      */
    public static void main(String [] args) {
        try {
            String registryServerName = args[0];
            new ChatApplication(registryServerName);
        } catch (Exception e) {
            System.err.println("Usage: java ChatApplication registryServerName [numberOfApplications]");
        }

    }
}


