// Copyright (C) 1999 Joshua Fox

package ddj;

import java.awt.*;
import java.awt.event.*;

/**
 * To achieve a modular design, you might want to structure your
 * system in as many as four layers:
 * (1) JSDT;
 * (2) The layer that deals with JSDT and excapsulates it from other layers;
 * (3) the business logic of your distributed collaboration system;
 * (4) the GUI layer.
 *
 * For an additional reduction of coupling, hide the layers from each other
 * with interfaces.
 * For a quicker solution, you might combine layers 2 and 3, as our "Chat"
 * example does.
 *
 * The principle of layer separation is demonstrated in this Chat
 * example by making sure that the Chat Application and Chat Server GUIs have
 * no references to JSDT. (This can be easily checked by examining the
 * import statements.)
 */
public class ChatApplicationGUI extends Frame implements Runnable {
    /**
     * The ChatApplication that provides the "business logic."
     */
    private ChatApplication chatApp;

    GridLayout chatGuiGridLayout = new GridLayout();

    Panel talkPanel = new Panel();
    BorderLayout talkPanelBorderLayout = new BorderLayout();
    TextArea receiveMessagesTextArea = new TextArea();

    Panel sendPanel = new Panel();
    BorderLayout sendPanelBorderLayout = new BorderLayout();
    Button sendButton = new Button ();
    TextField sendTextField = new TextField();

    Panel controlPanel = new Panel();
    GridLayout controlPanelGridLayout = new GridLayout();

    Panel clientPanel = new Panel();
    BorderLayout clientPanelBorderLayout = new BorderLayout();

    Panel clientListPanel = new Panel();
    BorderLayout clientListPanelBorderLayout = new BorderLayout();
    Label clientListLabel = new Label();
    List clientList = new List();


    Panel sessionPanel = new Panel();
    BorderLayout sessionPanelBorderLayout = new BorderLayout();
    List sessionList = new List();

    Panel clientNamePanel = new Panel();
    BorderLayout clientNamePanelBorderLayout = new BorderLayout();
    Label clientNameLabel = new Label();
    TextField clientNameTextField = new TextField();

    Panel sessionNamePanel = new Panel();
    BorderLayout sessionNamePanelBorderLayout = new BorderLayout();
    TextField sessionNameTextField = new TextField();
    Button disconnectButton = new Button();
    Button connectToSessionButton = new Button();

    Panel currentSessionPanel = new Panel();
    Label sessionLabel = new Label();
    TextField currentSessionTextField = new TextField();
    BorderLayout currentSessionBorderLayout = new BorderLayout();


    /**
     *
     */
    public ChatApplicationGUI(ChatApplication app) {
        chatApp = app;

        initialize();

        // A Thread to update the ChatGUI on the
        // Sessions currently in existence.
        Thread sessionMonitorThread =  new Thread(this);
        sessionMonitorThread.setName("SessionMonitorThread");
        sessionMonitorThread.setDaemon(true);
        sessionMonitorThread.start();
    }


   /**
    * @return the name which serves as the identifier of this client in the
    * JSDT system. The JSDT Client object encapsulates this name.
    * like to connect.
    */
    String getClientName() {
        return clientNameTextField.getText();
    }

    /**
     * @return the name of the session (conversation) to which the client would
     * like to connect.
     */
    String getSessionName(){
        return sessionNameTextField.getText();
    }

    /**
     * Messages that arrive from the JSDT come into the GUI through this method.
     * @param s the message that arrived.
     */
    void receiveMessage(String s) {
        receiveMessagesTextArea.append(s);
    }

    /**
     *
     */
    synchronized void setSessions(String [] sessions) {
        sessionList.removeAll();

        if (sessions==null) {
            return;
        }

        for (int i = 0; i< sessions.length; i++) {
            sessionList.addItem(sessions[i]);
        }
    }


    /**
     * This GUI uses Thread to monitor the Sessions currently in existence.
     * Port-binding conflicts make it difficult to use event-based monitoring
     * with a RegistryListener.
     */
     public void run() {
        while (true) {
            setSessions(chatApp.getCurrentSessions());

            try {
                Thread.sleep(4000);
            } catch (InterruptedException ie) {
            }
        }
    }

    /**
    * Set the list of clients in the current Session.
    */
    void setClientsInSession(String [] names) {
        clientList.removeAll();
        for (int i = 0; i < names.length; i++){
            clientList.add(names[i]);
        }
    }

    /**
     *
     */
     private void initialize()  {
        // to close the app:
        this.addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                chatApp.disconnect();
                dispose();
                System.exit(0);
            }
        });

        // layouts:
        chatGuiGridLayout.setRows(2);
        controlPanelGridLayout.setColumns(2);

        this.setLayout(chatGuiGridLayout);
        controlPanel.setLayout(controlPanelGridLayout);
        sendPanel.setLayout(sendPanelBorderLayout);
        clientNamePanel.setLayout(clientNamePanelBorderLayout);
        sessionPanel.setLayout(sessionPanelBorderLayout);
        clientPanel.setLayout(clientPanelBorderLayout);
        talkPanel.setLayout(talkPanelBorderLayout);
        sessionNamePanel.setLayout(sessionNamePanelBorderLayout);
        clientListPanel.setLayout(clientListPanelBorderLayout);
        currentSessionPanel.setLayout(currentSessionBorderLayout);

        // labels and text
        clientNameLabel.setText("Your name:");
        disconnectButton.setLabel("Disconnect From Session");
        connectToSessionButton.setLabel("Connect to Session");
        sessionLabel.setText("Your Session");
        clientListLabel.setText("Clients in your Session");

        this.setTitle("Chat");

        // colors:
        receiveMessagesTextArea.setForeground(Color.blue);
        receiveMessagesTextArea.setBackground(Color.yellow);
        clientList.setBackground(Color.yellow);
        currentSessionTextField.setBackground(Color.yellow);
        sessionList.setBackground(Color.yellow);
        sessionNameTextField.setBackground(Color.yellow);
        currentSessionPanel.setBackground(Color.lightGray);
        clientNamePanel.setBackground(Color.lightGray);
        clientListPanel.setBackground(Color.lightGray);

        // set params of GUI components:
        receiveMessagesTextArea.setEditable(false);
        currentSessionTextField.setEditable(false);
        sessionNameTextField.setEditable(false);
        sessionList.setMultipleMode(false);
        clientList.setMultipleMode(false);
        setEndSession();

        // Session name:

        sessionList.addActionListener(new java.awt.event.ActionListener() {
            //double-click
            public void actionPerformed(ActionEvent event) {
                sessionList_actionPerformed(event);
            }
        });

        sessionList.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(ItemEvent event) {
                sessionList_itemStateChanged(event);
            }
        });


        // Session dis/connect
        connectToSessionButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent event) {
              connectToSession();
            }
        });
        disconnectButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent event) {
                chatApp.disconnect();
                setEndSession();
            }
        });

        // to send text:
        sendButton.setLabel("Send");
        sendTextField.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent event) {
                sendTextField_actionPerformed(event);
            }
        });
        sendButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent event) {
                // the button does the same thing as the textField
                sendTextField_actionPerformed(event);
            }
        });


        // add components:
        this.add(talkPanel, null);
        talkPanel.add(sendPanel, BorderLayout.SOUTH);
        sendPanel.add(sendButton , BorderLayout.EAST);
        sendPanel.add(sendTextField, BorderLayout.CENTER);
        talkPanel.add(receiveMessagesTextArea, BorderLayout.CENTER);
        this.add(controlPanel, null);
        controlPanel.add(clientPanel, null);
        clientPanel.add(clientNamePanel, BorderLayout.SOUTH);
        clientNamePanel.add(clientNameLabel, BorderLayout.NORTH);
        clientNamePanel.add(clientNameTextField, BorderLayout.CENTER);
        clientPanel.add(clientListPanel, BorderLayout.CENTER);
        clientListPanel.add(clientListLabel, BorderLayout.NORTH);
        clientListPanel.add(clientList, BorderLayout.CENTER);
        controlPanel.add(sessionPanel, null);
        sessionPanel.add(sessionList, BorderLayout.CENTER);
        sessionPanel.add(sessionNamePanel, BorderLayout.SOUTH);
        sessionNamePanel.add(sessionNameTextField, BorderLayout.CENTER);
        sessionNamePanel.add(disconnectButton, BorderLayout.SOUTH);
        sessionNamePanel.add(connectToSessionButton, BorderLayout.NORTH);
        sessionPanel.add(currentSessionPanel, BorderLayout.NORTH);
        currentSessionPanel.add(sessionLabel, BorderLayout.WEST);
        currentSessionPanel.add(currentSessionTextField, BorderLayout.CENTER);

        this.setSize(new Dimension(600, 600));

        // to prevent overlap, offset windows
        int offset = (int) (150 * Math.random()) ;
        this.setLocation(10 + offset,
                         10 + (int)(offset/5));
        setVisible(true);
        clientNameTextField.requestFocus();

    }



    /**
     *
     */
    private void connectToSession() {
        if (clientNameTextField.getText().equals("") ||
            sessionNameTextField.getText().equals("") ) {
                new ComplaintDialog("Both Client and Session name must be set.");
                return;
        }


        try {
            clientNameTextField.setEnabled(false);
            chatApp.connectToSession();
            setStartSession();
        } catch (ChatException ce) {
            if (ce.getExceptionType() == ChatException.CLIENT_NAME_IN_USE) {
                clientNameTextField.setEnabled(true);
                new ComplaintDialog("Client name in use; choose another name.");
            } else {
                System.err.println("Exception in connecting to Session: " + ce);
                ce.printStackTrace();
            }
        }
    }

    /**
    *
    */
    void setStartSession() {
        sendButton.setEnabled(true);
        sendTextField.setEnabled(true);
        sendTextField.setBackground(Color.white);
        currentSessionTextField.setText(chatApp.getCurrentSession());
    }


    /**
    *
    */
    void setEndSession() {
        sendButton.setEnabled(false);
        sendTextField.setEnabled(false);
        sendTextField.setBackground(Color.darkGray);
        setClientsInSession(new String[0]);
        currentSessionTextField.setText("");
    }


    /**
     * Double-clicking the list of Sessions. Connect to session.
     */
    void sessionList_actionPerformed(ActionEvent actionEvent) {
        String sessionChosen = sessionList.getSelectedItem();
        sessionNameTextField.setText(sessionChosen);
        connectToSession();
    }

    /**
     * Single-clicking the list of Sessions. Put Session name in
     * textfield
     */
     void sessionList_itemStateChanged(ItemEvent event) {
        int clickedItem = ((Integer)event.getItem()).intValue();
        String s = sessionList.getItem(clickedItem);
        sessionNameTextField.setText(s);
    }


   /**
     *
     */
    void sendTextField_actionPerformed(ActionEvent actionEvent) {
        try {
            chatApp.sendMessage(sendTextField.getText());
            sendTextField.setText("");
        } catch (ChatException ce) {
            System.err.println("Exception in sending message " +
                 sendTextField.getText() + "; Exception " + ce);
        }
    }

    /**
     * Inner class
     */
    private class ComplaintDialog extends Dialog {
        /**
         *
         */
        ComplaintDialog(String s) {
            super(ChatApplicationGUI.this, "Click OK to Continue", true);
            setSize(300, 150);
            setLocation(500 , 200);
            Label label = new Label(s);
            label.setFont(new Font("Dialog", Font.PLAIN, 14));
            Button button = new Button("OK");
            button.setFont(new Font("Dialog", Font.PLAIN, 18));
            button.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                     dispose();
                }
            });

            addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent event) {
                     dispose();
                }
            });

            add(label, BorderLayout.CENTER);
            add(button, BorderLayout.SOUTH);


            setVisible(true);
        }
    } //end inner class
}
