// Copyright (C) 1999 Joshua Fox
package ddj;

import com.sun.media.jsdt.Channel;
import com.sun.media.jsdt.JSDTException;
import com.sun.media.jsdt.RegistryFactory;
import com.sun.media.jsdt.Session;
import com.sun.media.jsdt.SessionFactory;
import com.sun.media.jsdt.URLString;
import com.sun.media.jsdt.event.RegistryAdaptor;
import com.sun.media.jsdt.event.RegistryEvent;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.Properties;


/**
 * This class provides a server which starts some Sessions.
 * It reads the list of Sessions from a properties file,
 * whose name can be supplied on the command line
 * (default sessions.properties).
 */
public class DDJChatServer {
    private ChatServerGUI serverGui;
    private JSDTClient jsdtClient;
    private String sessionsFileName;
    private String localHostName;

    /**
     * Don't instantiate outside this class
     */
    private DDJChatServer(String fileName) {
        // The server must have a Client object
        // representing it in the JSDT system.
        jsdtClient = new JSDTClient("SERVER");
        sessionsFileName = fileName;
        serverGui = new ChatServerGUI(this);

        localHostName = "localhost";
        
        try {
            localHostName = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException uhe) {
            System.err.println("Unknown Host " + uhe);
        }

    }

    /**
     *
     */
    synchronized void go() {
        try {

            if (!RegistryFactory.registryExists(Constants.JSDT_IMPLEMENTATION)) {
                RegistryFactory.startRegistry(Constants.JSDT_IMPLEMENTATION);
            } else {
                System.err.println("Registry already exists.");
            }

            RegistryFactory.addRegistryListener(localHostName,
                                            Constants.JSDT_IMPLEMENTATION,
                                            new RegistryMonitor());

        } catch (JSDTException jsdte){
            System.err.println("Cannot start the registry; Exception " + jsdte);
        }

        startSessions();

    }

    /**
     * At start-up, start up the Sessions with names listed
     * in the properties file. The properties file is named
     * in the variable sessionsFileName, which
     * is "sessions.properties" by default, but which
     * may be specified on the command line.
     */
    private void startSessions() {
        Properties prop = new Properties();
        try {
            FileInputStream fis = new FileInputStream(sessionsFileName);
            BufferedInputStream bis = new BufferedInputStream(fis);
            prop.load(bis);
        } catch (FileNotFoundException fnfe) {
            System.err.println("File " + sessionsFileName + " not found.");
        } catch (Exception e) {
            System.err.println("Exception in loading " + sessionsFileName + " " + e);
        }


        Enumeration sessionNamesEnum = prop.propertyNames();
        while (sessionNamesEnum.hasMoreElements()) {
            String sessionName = (String)sessionNamesEnum.nextElement();
            startSessionIfNeeded(sessionName);
        }
    }


    /**
    *   Create the Session with the given name if none such exists.
    */
    void startSessionIfNeeded(String sessionName) {
       URLString sessionUrlString = URLString.createSessionURL(
                                            localHostName,
                                            Constants.SESSION_PORT,
                                            Constants.JSDT_IMPLEMENTATION,
                                            sessionName);

       boolean needToStart = true;

       try {

           needToStart = !SessionFactory.sessionExists(sessionUrlString);

       } catch (JSDTException jsdte) {

            System.err.println("Exception in determining if session " +

                            sessionUrlString + " exists: "  + jsdte);

       }


       if (needToStart) {

            startSession(sessionUrlString);

       }

    }




    /**
    *   Create the Session with the given URL
    */
    private synchronized void startSession(URLString sessionUrlString) {
        try {
            // The method createSession() creates the Session if
            // it does not exist, or gets a reference to it if
            // it does exist. In the client application, we will
            // not actually create the Session, just get a reference.
            // The "false" parameter in createSesssion() indicates
            // that we will create a Session, but not join it.
            Session session = SessionFactory.createSession(jsdtClient,
                                                          sessionUrlString,
                                                          false);
            Channel chatChannel = session.createChannel(jsdtClient,
                                                    Constants.CHAT_CHANNEL_NAME,
                                                    true, true, false);
         } catch (JSDTException jsdte) {
                System.err.println("JSDT Exception in startSession(" +
                        sessionUrlString + "): " + jsdte);
         }

    }

     /**
      *  End all Sessions
      */
    private synchronized void endAllSessions() {
        try {
            URLString [] sessions = SessionFactory.listSessions();
            for (int i = 0; i < sessions.length; i++ ) {
                endSession(sessions[i]);
            }
        } catch (JSDTException jsdte) {
            System.err.println("Exception in ending Sessions " + jsdte);
        }

    }


    /**
     * End the Session with the URL indicated by sessionUrlString
     */
    private void endSession(URLString sessionUrlString) {
       try {
          Session session = SessionFactory.createSession(jsdtClient,
                                                         sessionUrlString,
                                                         false);
          session.destroy(jsdtClient);
       } catch (JSDTException jsdte) {
          System.err.println("Exception in ending Session " +
                    sessionUrlString.getObjectName() + "; Exception " +  jsdte);
       }
    }


    /**
     * @return an array of the URLStrings all Sessions currently in the System,
     * (expressed as Strings, for the GUI to display).
     */
    public String [] getSessionUrls(){
        try {
            URLString[] sessionUrls = SessionFactory.listSessions();
            String [] sessionNames = new String[sessionUrls.length];
            // copy array
            for (int i =0; i < sessionUrls.length; i++){
                sessionNames[i] = sessionUrls[i].toString();
            }
            return sessionNames;
        } catch (JSDTException jsdte) {
            return new String[0];
        }


    }


     /**
      * End all Sessions and stop the Registry.
      */
    synchronized void stopServer() {
        try {
            endAllSessions();
            RegistryFactory.stopRegistry(Constants.JSDT_IMPLEMENTATION);
        } catch (JSDTException jsdte) {
            System.err.println("Cannot stop the registry; Exception " +jsdte);
        }
    }

     /**
      * @param args: if args[0] is given, its is used as the name of the
      * file with the name of Sessions to start up. The default
      * is sessions.properties
      */
    public static void main(String [] args) {
        String fileName = "sessions.properties";
        if (args.length > 0) {
           fileName = args[0];
        }

        DDJChatServer server = new DDJChatServer(fileName);
        server.go();
    }



    /**
    * Inner class that monitors when Sessions are added
    */
    private class RegistryMonitor extends RegistryAdaptor {
      /**
       * Call-back method; called when a Session is created.
       */
      public void sessionCreated(RegistryEvent event) {
            serverGui.addSession(event.getResourceName().toString());
      }
   }
}



