// ----------- interp.cpp
#include <stdio.h>
#include <process.h>
#include "interp.h"

// namespace DDJScriptInterpreter	{

// --------------- tokens
const token AUTOINC    = 'P';
const token AUTODEC    = 'D';
const token EQUALTO    = 'E';
const token NOTEQUAL   = 'N';
const token GE         = 'G';
const token LE         = 'L';
const token IF         = 'f';
const token ELSE       = 'e';
const token WHILE      = 'w';
const token FOR        = 'F';
const token CHAR       = 'c';
const token INT        = 'i';
const token STRING     = 's';
const token PLUS       = '+';
const token MINUS      = '-';
const token MULTIPLY   = '*';
const token DIVIDE     = '/';
const token EQUAL      = '=';
const token LESS       = '<';
const token GREATER    = '>';
const token NOT        = '!';
const token LPAREN     = '(';
const token RPAREN     = ')';
const token LBRACE     = '{';
const token RBRACE     = '}';
const token LBRACKET   = '[';
const token RBRACKET   = ']';
const token COMMA      = ',';
const token AND        = '&';
const token OR         = '|';
const token QUOTES     = '"';
const token QUOTE      = '\'';
const token UNDERSCORE = '_';
const token SEMICOLON  = ';';
const token IDENT      = 'I';
const token CONSTANT   = 'C';
const token LINENO     = 127;
const token RETURN     = 'r';
const token BREAK      = 'b';

// -------- table of key words and their tokens ---------
SInterpreter::Keyword SInterpreter::keywords[] = {
	Keyword("for",   Token(FOR)),
	Keyword("while", Token(WHILE)),
	Keyword("if",    Token(IF)),
	Keyword("else",  Token(ELSE)),
	Keyword("int",   Token(INT)),
	Keyword("char",  Token(CHAR)),
	Keyword("string",Token(STRING)),
	Keyword("return",Token(RETURN)),
	Keyword("break", Token(BREAK)),
	Keyword("",      Token(0))
};
// ------------ table of direct translate tokens --------
token SInterpreter::tokentbl[] = {
	COMMA,
	LBRACE,
	RBRACE,
	LPAREN,
	RPAREN,
	EQUAL,
	NOT,
	LESS,
	GREATER,
	AND,
	OR,
	QUOTES,
	SEMICOLON,
	LBRACKET,
	RBRACKET,
	MULTIPLY,
	DIVIDE,
	PLUS,
	MINUS,
	EOF,
	LINENO,
	0
};

SInterpreter::SInterpreter(const Intrinsic* inf)
{
	// --- add intrinsic functions to the symbol table --- 
	while (inf->fn != 0)	{
		Symbol sym(ifunction);
		std::string name;
		const std::string& sig = inf->signature;
		if (sig.substr(0, 7) == "string ")	{
			sym.datum.type = strng;
			name = sig.substr(7, sig.length()-7);
		}
		else	{
			sym.datum.type = number;
			if (sig.substr(0, 4) == "int ")
				name = sig.substr(4, sig.length()-4);
			else if (sig.substr(0, 5) == "char ")
				name = sig.substr(5, sig.length()-5);
			else
				name = sig;
		}
		sym.identifier = addidentifier(name);
		sym.fn = inf->fn;
		symboltable.push_back(sym);
		inf++;
	}
	initialize();
}

void SInterpreter::initialize()
{
	skipping = 0;
	linenumber = 1;
	currentscope = 0;
	breaking = false;
	returning = false;
	scanned = false;
	tokens.clear();
}

#ifdef _DEBUG
FILE* fp;
#endif

int SInterpreter::interpret()
{
#ifdef _DEBUG
	fp = fopen("debug.txt", "wt");
#endif
	initialize();
	lexicalscan();
	scanned = true;
	tokiter = tokens.begin();
	currentscope = symboltable.size();
	declarators(false);

#ifdef _DEBUG
	if (fp)	{
		fclose(fp);
		fp = 0;
	}
#endif

	int ndx;
	Symbol sym;
	bool foundmain;
	if ((foundmain = findsymbol(ndx, "main")) == true)	{
		sym = symboltable[ndx];
		if (sym.type != pfunction)
			foundmain = false;;
	}
	if (!foundmain)
		throw SIException(NOMAIN, 0);
	tokiter = tokens.begin() + sym.entry;	// function entry point
	needtoken(LPAREN);
	needtoken(RPAREN);
	needtoken(LBRACE);
	compound_statement(symboltable.size());   // execute the main function

	// global declarations (except intrinsic functions) go out of scope
	outofscope();
	return frtn.value;
}

void SInterpreter::outofscope()
{
	if (currentscope < symboltable.size())
		symboltable.erase(symboltable.begin()+currentscope, symboltable.end());
}

// -- convert a script program to tokens for interpreter
void SInterpreter::lexicalscan()
{
	Token tk;

	do	{
		tk = compilenextsourcetoken();
		tk.tokennumber = tokens.size() + 1;
		tokens.push_back(tk);
	} while (tk.tok != EOF);

#ifdef _DEBUG
	// --- testing: dump the token buffer
	if (fp != 0)	{
		for (tokiter = tokens.begin(); tokiter != tokens.end(); tokiter++)	{
			Token tk = *tokiter;
			fprintf(fp, "\n%d: %c (%02x) %d %s", 
				tk.tokennumber,
				tk.tok, tk.tok, tk.datum.value, tk.datum.strval.c_str());
		}
	}
#endif
}

// -- compile the next token from the source code
Token SInterpreter::compilenextsourcetoken()
{
	unsigned int c;
	Token tk;

	std::string word;

	do                 // bypass white space
		c = getsourcechar();
	while (iswhite(c));

	switch (c)	{

		case '\n':
			tk.tok = LINENO;
			break;
	
		case QUOTE:
			tk.tok = CONSTANT;              // quoted constant ('x') 
			if ((c = getrawsourcechar()) == '\\')   // escape sequence (\n)  
				c = escseq();
		 	tk.datum = Datum(c);
			if (getrawsourcechar() != QUOTE)     // needs the other quote    
				throw SIException(MISSING, LineNumber(), "double quote");
			break;
	
		case QUOTES:
			tk.tok = STRING;                 // quoted string "abc"  
			while ((c = getrawsourcechar()) != QUOTES)
				word += (c == '\\' ? escseq() : c);
			tk.datum = Datum(word);
			break;
	
		default:
			if (isidentchar(c))	{
				do	{
					word += c;
					c = getsourcechar();
				} while (isidentchar(c));
				ungetsource(c);
			}
			else
				word = c;
			break;
	}

	if (tk.tok == 0 && word.length() == 1)	       {
		// --- check for an operator, paren, etc.
		int t2;
		token tkn = word[0];
		int tbllen = sizeof tokentbl / sizeof tokentbl[0];
		int i;
		for (i = 0; i < tbllen; i++)	{
			if (tkn == tokentbl[i])	{
				switch (tkn)  {
					case EOF:
						break;
					case AND:
						if ((t2 = getrawsourcechar()) != AND)
							throw SIException(MISSING, LineNumber(), "&");
						break;
					case OR:
						if (getrawsourcechar() != OR)
							throw SIException(MISSING, LineNumber(), "|");
						break;
					case PLUS:
					case MINUS:
						if ((t2 = getrawsourcechar()) == tkn)
							tkn = tkn == PLUS ? AUTOINC : AUTODEC;
						else
							ungetsource(t2);
						break;
					default:
						t2 = getrawsourcechar();
						if (t2 == EQUAL)    {
							switch (tkn)  {
								case EQUAL:
									tkn = EQUALTO;
									break;
	   	   	   	   	   	   	   	case NOT:
									tkn = NOTEQUAL;
									break;
	   	   	   	   	   	   	   	case LESS:
									tkn = LE;
									break;
	   	   	   	   	   	   	   	case GREATER:
									tkn = GE;
									break;
								default:
									ungetsource(t2);
									break;
							}
						}
						else
							ungetsource(t2);
						break;
				}
				tk.tok = tkn;
				break;
			}
		}	 
	}

	if (tk.tok == 0)	     {
		// --- check for a keyword
		struct Keyword* k = keywords;
		while (k->kw.length() != 0)	{
			if (k->kw == word)	{
				tk = k->kwtoken;
				break;
			}
			k++;
		}
	}

	if (tk.tok == 0)	 {
		// --- check for an identifier
		if (isalpha(word[0]) || word[0] == UNDERSCORE)	{
			for (int i = 1; i < word.length(); i++)
				if (!isidentchar(word[i]))
					throw SIException(NOTIDENT, LineNumber());
			tk.tok = IDENT;
			tk.datum.value = addidentifier(word);
		}
	}

	if (tk.tok == 0)	{
		// --- check for decimal, hex or octal constant
		if (word[0] == '0')	{
			if (tolower(word[1]) == 'x')	{
				// --- hex constant
				int w = 2;
				int n = 0;
				while (w < word.length()) {
					char ch = tolower(word[w++]);
					if (!isdigit(ch) || ch < 'a' || ch > 'f')
						throw SIException(INVALIDCONSTANT, LineNumber());
					n = (n*16)+(isdigit(ch) ? ch-'0' : ch - 'a' + 10);
				}
				tk.tok = CONSTANT;
				tk.datum = Datum(n); // converted hex constant 
			}
			else	{
				// --- octal constant
				int w = 1;
				int n = 0;
				while (w < word.length()) {
					char ch = word[w++];
					if (ch < '0' || ch > '7')
						throw SIException(INVALIDCONSTANT, LineNumber());
					n = (n*8) + ch - '0';
				}
				tk.tok = CONSTANT;
				tk.datum = Datum(n); // converted hex constant 
			}
		}
		else if (isdigit(word[0]))	{
			// --- decimal constant
			int w = 0;
			int n = 0;
			while (w < word.length()) {
				char ch = word[w++];
				if (ch < '0' || ch > '9')
					throw SIException(INVALIDCONSTANT, LineNumber());
				n = (n*10) + ch - '0';
			}
			tk.tok = CONSTANT;
			tk.datum = Datum(n); // converted hex constant 
		}
	}

	if (tk.tok == 0)
		throw SIException(UNRECOGNIZED, LineNumber());
	return tk;
}

// ---- escape sequence in literal constant or string ---- 
int SInterpreter::escseq()
{
	int c = getrawsourcechar();
	return (c == 'n' ? '\n' :
			c == 't' ? '\t' :
			c == 'f' ? '\f' :
			c == 'a' ? '\a' :
			c == 'b' ? '\b' :
			c == 'r' ? '\r' :
			c == '0' ? '\0' : c);
}
// ------- get a character from the input stream -------- 
int SInterpreter::getsourcechar()
{
	int c;
	if ((c = getsource()) == '/')   {
		int c1;
		if ((c1 = getsource()) == '/')	{
			while ((c = getsource()) != '\n' && c != EOF)
				;
		}
		else
			ungetsource(c1);	// divide not comment
	}
	return c;
}
// ------- read a character from input, error if EOF ------ 
int SInterpreter::getrawsourcechar()
{
	int c;
	if ((c = getsource()) == EOF)
		throw SIException(EARLYEOF, LineNumber());
	if (c == '\n')
		linenumber++;
	return c;
}
// --------- build the symbol table --------- 
void SInterpreter::declarators(bool islocal)
{
#ifdef _DEBUG
	int top = symboltable.size();
#endif

	while (declarator(islocal))
		;

#ifdef _DEBUG
	// --- testing: dump the symbol table
	if (fp != 0 && top != symboltable.size())	{
		fprintf(fp,"\n--------------------symbols----------------");
		for (symbol_iter siter = symboltable.begin()+top; siter != symboltable.end(); siter++)	{
			Symbol sym = *siter;
			fprintf(fp,"\n%d %s", sym.type, identifiers[sym.identifier].c_str());
		}
	}
#endif
}

bool SInterpreter::declarator(bool islocal, bool isparameter)
{
	Token tk = nexttoken();
	bool isstring = false;
	switch (tk.tok)  {
		case STRING:
			isstring = true;
		case CHAR:
		case INT:
		{
			bool iscomma = false;
			bool isfunc = false;
			do	{
				tk = needtoken(IDENT);
				Symbol sym;
				sym.identifier = tk.datum.value;
				int ndx;

				if (findsymbol(ndx, sym.identifier, islocal ? currentscope : 0) == true)
					throw SIException(DUPL_DECLARE, LineNumber(), identifiers[sym.identifier]);

				if (!islocal && istoken(LPAREN))	{
					// declares a function
					if (iscomma)
						throw SIException(SYNTAX, LineNumber());
					sym.type = pfunction;
					sym.datum.type = isstring ? strng : number;
					prevtoken();
					sym.entry = tokiter - tokens.begin();
					skippair(LPAREN, RPAREN);
					skippair(LBRACE, RBRACE);
					isfunc = true;
				}
				else	{
					// declares a variable
					sym.type = variable;
					if (!isparameter)	{
						if (istoken(EQUAL))
							sym.datum = expression();
						iscomma = istoken(COMMA);
					}
				}
				symboltable.push_back(sym);
			} while (iscomma);
			if (!isparameter && !isfunc)
				needtoken(SEMICOLON);
			return true;
		}
		default:
			prevtoken();
			break;
	}
	return false;
}

// --------- a function is called ---------- 
Datum SInterpreter::function(Symbol sym)
{
	typedef std::vector<Datum> Parameters;
	Parameters params;
	needtoken(LPAREN);

	frtn = Datum(0);				// global function return value

	token_iter citer = tokiter;		// for caller's line number for mismatched argument exception

	while (!istoken(RPAREN))	{
		Datum param = expression();
		params.push_back(param);	// build params
		if (!istoken(COMMA))    {	// into parameter vector
			needtoken(RPAREN);
			break;
		}
	}

	switch (sym.type)	 {
		case ifunction:
			if (!skipping)	{
				std::vector<int> pars;
				Parameters::iterator iter;
				int i = 0;
				for (iter = params.begin(); iter != params.end(); iter++)	{
					Datum& d = *iter;
					if (d.type == strng)
						pars.push_back(reinterpret_cast<int>(d.strval.c_str()));
					else
						pars.push_back(d.value);
				}
				// call intrinsic func
				frtn.value = (*sym.fn)(&pars[0]);
				if (sym.datum.type == strng)	{
					frtn.type = strng;
					frtn.strval = std::string(reinterpret_cast<char*>(frtn.value));
					frtn.value = 0;
				}
			}
			break;
		case pfunction:
		{
			// save previous context
			token_iter saveiter = tokiter;
			// execute the function
			tokiter = tokens.begin() + sym.entry;	// function entry point
			int savescope = currentscope;
			currentscope = symboltable.size();
			needtoken(LPAREN);
			int pct = params.size();
			int i = 0;
			while (i < pct) {
	   	   	    // parameters into symbol tbl
				if (!declarator(true, true))
					break;
				// arguments into parameters
				symboltable.back().datum = params[i++];
				if (i < pct)
					istoken(COMMA);	// bypass a comma if it exists
			}
			if (istoken(COMMA) || i < pct)	{
				tokiter = citer;		// for caller's line number
				throw SIException(MISMATCHEDARG, LineNumber());
			}
			needtoken(RPAREN);
			needtoken(LBRACE);
			compound_statement(currentscope);   // execute the function
			currentscope = savescope;
			breaking = returning = false;
			// restore caller's context
			tokiter = saveiter;		// return to caller
			break;
		}
		default:
			throw SIException(NOTFUNC, LineNumber());
			break;
	}
	return frtn;              // the function's return value
}

// ------- execute one statement or a {} block -------- 
void SInterpreter::statements()
{
	if (istoken(LBRACE))
		compound_statement(symboltable.size());
	else
		statement();
}
// -------- execute a {} statement block ----------- 
void SInterpreter::compound_statement(int scope)
{
	token_iter saveiter = tokiter;
	int savescope = currentscope;
	currentscope = scope;
	declarators(true);
	while (!istoken(RBRACE) && !breaking && !returning)
		statements();
	if (breaking || returning)	{
		tokiter = saveiter;
		prevtoken();	// point to the opening { brace
		skippair(LBRACE, RBRACE);	// skip to end of block
	}
	outofscope();	// local declarations go out of scope
	currentscope = savescope;
}
// --------- execute a single statement ---------- 
void SInterpreter::statement()
{
	token_iter fortest;
	token_iter forloop;
	token_iter forblock;

	token_iter saveiter = tokiter;
	   
	Datum rtn;

	Token tk = nexttoken();
	switch (tk.tok)    {
		case IF:
			needtoken(LPAREN);
			rtn = expression();		// condition being tested
			needtoken(RPAREN);
			if (rtn.value == 0)
				statements();		// condition is true 
			else
				skip_statements();	// condition is false
			while (istoken(ELSE))
				if (rtn.value != 0)	// do the reverse for else
					skip_statements();
				else
					statements();
			break;
		case WHILE:
			breaking = returning = false;
			saveiter = tokiter;
			do	{
				tokiter = saveiter;
				needtoken(LPAREN);
				rtn = expression(); // the condition tested
				needtoken(RPAREN);
				if (rtn.value != 0)
					statements();      // true
				else
					skip_statements(); // false
			} while (rtn.value != 0 && !breaking && !returning);
			breaking = returning = false;
			break;
		case FOR:
		{
			needtoken(LPAREN);
			if (!istoken(SEMICOLON))    {
				expression();  // initial expression
				needtoken(SEMICOLON);
			}
			fortest = tokiter;   // fortest -> terminating test
			tokiter = saveiter;	 // point back to for-expression
			nexttoken();
			skippair(LPAREN,RPAREN);
			forblock = tokiter;  // forblock -> block to run
			tokiter = fortest;	 // point back to terminating test
			breaking = returning = false;
			bool looping = true;
			do	{
				if (!istoken(SEMICOLON))    {
					Datum d = expression();
					if (d.type == strng)
						throw SIException(SYNTAX, LineNumber());
					looping = !!d; // terminating test
					needtoken(SEMICOLON);
				}
				if (looping)	{
					forloop = tokiter;
					tokiter = forblock;
					statements();     // the loop statement(s)
					looping = !breaking && !returning;
					if (looping)	{
						tokiter = forloop;
						if (!istoken(RPAREN))   {
							expression();  // the end of loop expr
							needtoken(RPAREN);
						}
						tokiter = fortest;
					}
				}
			} while (looping);
			tokiter = forblock;
			skip_statements();      // skip past the block
			breaking = returning = false;
			break;
		}
		case RETURN:
			if (!istoken(SEMICOLON))    {
				Datum rtn = expression();
				if (!skipping)
					frtn = rtn;
				needtoken(SEMICOLON);
			}
			returning = skipping == 0;
			break;
		case BREAK:
			needtoken(SEMICOLON);
			breaking = skipping == 0;
			break;
		case IDENT:
			prevtoken();
			expression();
			needtoken(SEMICOLON);
			break;
		default:
		{
			prevtoken();
			std::string msg;
			msg += tokiter->tok;
			throw SIException(OUTOFPLACE, LineNumber(), msg);
		}
	}
}
// -------- bypass statement(s) ------------ 
void SInterpreter::skip_statements()
{
	if (istoken(LBRACE))	{
		prevtoken();
		skippair(LBRACE, RBRACE);	// skip to end of block
	}
	else	{
		skipping++;		// semaphore that suppresses assignments,
		statements();	// breaks,returns,++,--,function calls
		--skipping;   	// turn off semaphore
	}
}
// -------- recursive descent expression analyzer -------- 
Datum SInterpreter::primary()
{
	Datum rtn;

	Token tk = nexttoken();
	switch (tk.tok)  {
		case LPAREN:
			rtn = expression();
			needtoken(RPAREN);
			break;
		case NOT:
			rtn = Datum(!primary());
			break;
		case MINUS:
			rtn = -primary();
			break;
		case CONSTANT:
			rtn = tk.datum;
			break;
		case STRING:
			rtn = tk.datum;
			break;
		case AUTOINC:
		case AUTODEC:
			needtoken(IDENT);
		case IDENT:
		{
			int ndx;
			if (findsymbol(ndx, tk.datum.value) == false)
				throw SIException(UNDECLARED, LineNumber(), identifiers[tk.datum.value]);

			Symbol& sym = symboltable[ndx];
			if (sym.type == pfunction || sym.type == ifunction)  {
				// ---- this is a function call ---- 
				Datum r = function(sym);
				if (!skipping)
					rtn = r;
				break;
			}
			// dereferencing a variable
			bool autoinc = false, autodec = false;
			if (sym.datum.type != strng)  {
				if (!skipping && tk.tok == AUTOINC)
					++(sym.datum.value);
				else if (!skipping && tk.tok == AUTODEC)
					--(sym.datum.value);
				autoinc = istoken(AUTOINC);
				autodec = istoken(AUTODEC);
			}
			// --- can't use sym from here down. It might become invalid after expression()
			if (istoken(EQUAL))   {
				rtn = expression();
				if (!skipping)
					symboltable[ndx].datum = rtn;
			}
			rtn = symboltable[ndx].datum;
			if (!skipping)	{
				if (autoinc)
					symboltable[ndx].datum.value++;
				if (autodec)
					symboltable[ndx].datum.value--;
			}
			break;
		}
		default:
		{
			std::string msg;
			msg += tk.tok;
			throw SIException(OUTOFPLACE, LineNumber(), msg);
		}
	}
	return rtn;
}
Datum SInterpreter::mult()
{
	Datum drtn;
	Datum rtn = primary();
	while (true)
		if (istoken(MULTIPLY))	{
			drtn = primary();
			rtn = rtn * drtn;
		}
		else if (istoken(DIVIDE))   {
			drtn = primary();
			if (drtn.type == number && drtn.value == 0)
				throw SIException(DIVIDEERR, LineNumber());
			rtn = rtn / drtn;
		}
		else
			break;
	return rtn;
}
Datum SInterpreter::plus()
{
	Datum rtn = mult();
	while (true)
		if (istoken(PLUS))
			rtn = rtn + mult();
		else if (istoken(MINUS))
			rtn = rtn - mult();
		else
			break;
	return rtn;
}
Datum SInterpreter::le()
{
	Datum rtn = plus();
	while (true)
		if (istoken(LE))
			rtn = Datum(rtn <= plus());
		else if (istoken(GE))
			rtn = Datum(rtn >= plus());
		else if (istoken(LESS))
			rtn = Datum(rtn < plus());
		else if (istoken(GREATER))
			rtn = Datum(rtn > plus());
		else
			break;
	return rtn;
}
Datum SInterpreter::eq()
{
	Datum rtn = le();
	while (true)
		if (istoken(EQUALTO))
			rtn = Datum(rtn == le());
		else if (istoken(NOTEQUAL))
			rtn = Datum(rtn != le());
		else
			break;
	return rtn;
}
Datum SInterpreter::and()
{
	Datum rtn = eq();
	while (istoken(AND))
		rtn = Datum(eq() && rtn);
	return rtn;
}
Datum SInterpreter::expression()
{
	try	{
		Datum rtn = and();
		while (istoken(OR))
			rtn = Datum(and() || rtn);
		return rtn;
	}
	catch (SIException sex)	{
		if (sex.lineno == 0)
			sex.lineno = LineNumber();
		throw sex;
	}
}
// ----- skip the tokens between a matched pair ----- 
void SInterpreter::skippair(token ltkn, token rtkn)
{
	int pairct = 0;
	Token tk = nexttoken();
	if (tk.tok != ltkn)
		throw SIException(ltkn == LBRACE ? BRACERR : PARENERR, LineNumber());
	while (true)    {
		if (tk.tok == ltkn)
			pairct++;
		if (tk.tok == rtkn)
			if (--pairct == 0)
				break;
		tk = nexttoken();
		if (tk.tok == EOF)
			throw SIException(ltkn == LBRACE ? BRACERR : PARENERR, LineNumber());
	}
}
// ----- a specified token is required next ----- 
Token SInterpreter::needtoken(token tkn)
{
	Token tk = nexttoken();
	if (tk.tok != tkn)	{
		char m[] = "x";
		m[0] = static_cast<char>(tkn);
		std::string t(m);
		throw SIException(MISSING, LineNumber(), t);
	}
	return tk;
}
// ----- test for a specifed token next in line ----- 
bool SInterpreter::istoken(token tkn)
{
	Token tk = nexttoken();
	if (tk.tok != tkn)
		prevtoken();;
	return tk.tok == tkn;
}
// ----- get the next token from the buffer ----- 
Token SInterpreter::nexttoken()
{
	while (tokiter->tok == LINENO)
		tokiter++;
	return *tokiter++;
}
// ----- get the next token from the buffer ----- 
void SInterpreter::prevtoken()
{
	do
		--tokiter;
	while (tokiter->tok == LINENO);
}
// --- Add an identifer to the table if not already in the table. Return its index
int SInterpreter::addidentifier(const std::string& name)
{
	int i;
	for (i = 0; i < identifiers.size(); i++)
		if (identifiers[i] == name)
			return i;
	identifiers.push_back(name);
	return i;	
}

// --- find a symbol on the symbol table by text name (called only for "main")
bool SInterpreter::findsymbol(int& ndx, const std::string& name)
{
	for (int i = 0; i < identifiers.size(); i++)
		if (identifiers[i] == name)
			return findsymbol(ndx, i);
	return false;
}

// --------- find a symbol on the symbol table by identifier offset --------- 
bool SInterpreter::findsymbol(int& ndx, int nameoffset, int fromscope)
{
	bool rtn = false;
	if (!symboltable.empty() && fromscope < symboltable.size())	{
		symbol_iter siter = symboltable.end();
		ndx = symboltable.size();
		do	{
			Symbol sym = *(siter-1);
			--ndx;
			if (sym.identifier == nameoffset)	{
				rtn = true;
				break;
			}
			--siter;
		} while (siter != symboltable.begin() + fromscope);
	}
	return rtn;
}

int SInterpreter::LineNumber()
{
	if (scanned)	{
		linenumber = 1;
		for (token_iter ti = tokens.begin(); ti < tokiter; ti++)
			if (ti->tok == LINENO)
				linenumber++;
	}
	return linenumber;
}


// } // namespace DDJScriptInterpreter
